@@ -1,5 +1,5 @@
 package Git::Raw::Index;
-$Git::Raw::Index::VERSION = '0.75';
+$Git::Raw::Index::VERSION = '0.76';
 use strict;
 use warnings;
 
@@ -11,7 +11,7 @@ Git::Raw::Index - Git index class
 
 =head1 VERSION
 
-version 0.75
+version 0.76
 
 =head1 DESCRIPTION
 
@@ -35,11 +35,12 @@ Retrieve the L<Git::Raw::Repository> owning the index.
 Add C<$entry> to the index. C<$entry> should either be the path of a file
 or alternatively a L<Git::Raw::Index::Entry>.
 
-=head2 add_frombuffer( $path, $buffer )
+=head2 add_frombuffer( $path, $buffer, [$mode] )
 
 Add or update an entry from an in memory file. The entry will be placed at C<$path>
 with the contents of C<$buffer>. C<$buffer> may either be string or a reference
-to a string. Returns a L<Git::Raw::Index::Entry> object.
+to a string. C<$mode> is the file mode; it defaults to C<0100644>. Returns a
+L<Git::Raw::Index::Entry> object.
 
 =head2 add_all( \%opts )
 
@@ -192,6 +193,10 @@ The name of the common ancestor side of conflicts.
 Specify content automerging behaviour. Valid values are C<"ours">, C<"theirs">,
 and C<"union">.
 
+=item * "marker_size"
+
+The size of conflict markers.
+
 =item * "flags"
 
 Merge file flags. Valid values include:
